import sqlite3

from sqlite3 import Error, Connection
from sqlite3.dbapi2 import Cursor

import logging as log

# import re
import random

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Tuple


def processwords(text: str) -> List[Tuple[str, str]]:

    # words are just split by whitespaces. This means that punctuation is included in words.
    words: List[Any] = text.split(" ")

    # Add a NONE at the end of the list, to mark the end of a message AND the beginning of one.
    words.append(None)
    words.insert(0, None)

    return list(zip(words, words[1:]))


def makemarkov(worddict: Dict[str | None, List[Tuple[str, int]]]) -> str:

    # curword = random.choice(list(worddict.keys()))

    # Initial value is None. This value is used both to mark the end of a sentence and to mark the beginning of one.
    curword = None

    res = ""

    # implement basic memoization for frequency dict

    mem = {}

    while True:  # We just break when curword is NONE again.

        # build frequency dict if not memoize
        if curword not in mem:
            wordlist = worddict[curword]
            l = len(wordlist)
            candidates = []
            weights_list = []
            sum = 0

            for word, weight in wordlist:
                candidates.append(word)
                weights_list.append(weight)
                sum += weight

            weights = list(map(lambda x: x / l, weights_list))

            # memoize candidates and weights

            mem[curword] = (candidates, weights)

        # if they're memoize, pick random
        candidates, weights = mem[curword]

        pick = random.choices(candidates, weights)

        curword = pick[0]

        if curword is None:
            break
        else:
            res += " " + curword

    return res


class Markov(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def store(self, text: str, grp: int):
        """Store a message in a database. The telegram grpup id is used to identify the group."""
        pass

    @abstractmethod
    def generate(self, grp: int) -> str:
        """Generate a random message from said group."""
        pass

    @abstractmethod 
    def delete(self, grp: int):
        """Deletes all messages from group"""
        pass

    @abstractmethod
    def status(self, grp: int) -> str:
        """Print overall status of the database."""
        pass


class MarkovLite(Markov):
    def __init__(self, db_file="markov.db", inmemory=False):
        """Initialize sqlite3 backend markov object."""
        log.info(sqlite3.version)

        self.inmemory = inmemory

        conn = self.create_connection(db_file)

        if conn == None:
            log.error("Could not connect to database.")
            raise ValueError

        self.conn: Connection = conn

        self.create_tables()

    def store(self, text: str, grp: int):
        log.debug(f"storing {text} from group {grp}")

        cursor = self.conn.cursor()
        
        try:
            for word_pair in processwords(text):
                self.insert_words(cursor, word_pair, grp)
        except Exception as e:
            log.debug(f"Something went wrong: {e}")

        log.debug("committing changes...")

        self.conn.commit()

    def generate(self, grp: int) -> str:
        log.debug(f"generated a string from group : {grp}")

        worddict = self.get_worddict(grp)

        return makemarkov(worddict)

    def delete(self, grp: int):

        log.info(f"Deleting all words for group {grp}")

        sql = f"""DELETE FROM words WHERE group_id={grp}"""

        cursor = self.conn.cursor() 

        cursor.execute(sql)

        log.info(f"Words deleted.")



    def status(self, grp: int) -> str:

        words = self.get_worddict(grp)

        total = sum(
            [
                p[1] if p[0] is not None else 0
                for wordlist in words.values()
                for p in wordlist
            ]
        )

        res = f"""Database type: {"in-memory" if self.inmemory else "SQLite3 file"} 
            
            Logged a total of:
            - {len(words)} word pairs 
            - {len(words[None]) - 1 } total messages 
            - {total} total words
            """

        return res

    def create_connection(self, db_file: str) -> Connection | None:
        """create a database connection to a SQLite database"""
        conn = None
        try:
            if self.inmemory:
                conn = sqlite3.connect(":memory:")
            else:
                conn = sqlite3.connect(db_file)
        except Error as e:
            log.error(e)
            return None
        finally:
            return conn

    def create_tables(self):
        """create a words table"""
        cursor = self.conn.cursor()

        words_table = f"""CREATE TABLE IF NOT EXISTS words (
            group_id integer NOT NULL,
            word text,
            next_word text,
            count integer NOT NULL,
            WITHOUT ROWID,
            PRIMARY KEY (group_id, word, next_word) 
            );
        """

        try:

            log.info("Creating words tables into databse")

            cursor.execute(words_table)

            log.debug("committing changes...")
            self.conn.commit()
        except Error as e:
            log.error("Could not create table:")
            log.error(e)

    def insert_words(self, cursor: Cursor, word_pair: Tuple[str, str], grp: int):
        """Insert a pair of words into the database."""

        log.debug(f"Inserting words {word_pair} for group with id {grp}")

        sql = """INSERT INTO words(group_id, word, next_word, count)
              VALUES(?,?,?,?) ON CONFLICT(group_id, word, next_word) DO UPDATE SET count = count + 1 
            """

        parameters = (grp, word_pair[0], word_pair[1], 1)

        cursor.execute(sql, parameters)

    def get_worddict(self, grp: int) -> Dict[str | None, List[Tuple[str, int]]]:
        """Get from the database a dictionary of words."""

        cursor = self.conn.cursor()

        cursor.execute(f"SELECT * FROM words WHERE group_id={grp}")

        worddict = {}
        for _, word, nextword, count, _ in cursor.fetchall():
            if not word in worddict:
                worddict[word] = [(nextword, count)]
            else:
                worddict[word] += [(nextword, count)]

        return worddict


class MarkovPost(Markov):
    def __init__(self):
        pass
